@@ -29,28 +29,71 @@
 #include <pybind11/functional.h>
 #include <pybind11/pybind11.h>
 #include <pybind11/stl.h>
+
 #include <functional>
 #include <map>
 
 namespace py = pybind11;
 
 struct xatmi_exception : public std::exception {
- public:
-  explicit xatmi_exception(int code) : code(code), message(tpstrerror(code)) {}
-  int code;
-  std::string message;
+ private:
+  int code_;
+  std::string message_;
 
-  const char *what() const noexcept override { return message.c_str(); }
+ public:
+  explicit xatmi_exception(int code)
+      : code_(code), message_(tpstrerror(code)) {}
+
+  const char *what() const noexcept override { return message_.c_str(); }
+  int code() const noexcept { return code_; }
 };
 
 struct fml32_exception : public std::exception {
-  explicit fml32_exception(int code) : code(code), message(Fstrerror32(code)) {}
-  int code;
-  std::string message;
+ private:
+  int code_;
+  std::string message_;
 
-  const char *what() const noexcept override { return message.c_str(); }
+ public:
+  explicit fml32_exception(int code)
+      : code_(code), message_(Fstrerror32(code)) {}
+
+  const char *what() const noexcept override { return message_.c_str(); }
+  int code() const noexcept { return code_; }
 };
 
+static PyObject *TuxedoException_tp_str(PyObject *selfPtr) {
+  py::str ret;
+  try {
+    py::handle self(selfPtr);
+    py::tuple args = self.attr("args");
+    ret = py::str(args[0]);
+  } catch (py::error_already_set &e) {
+    ret = "";
+  }
+
+  ret.inc_ref();
+  return ret.ptr();
+}
+
+static PyObject *TuxedoException_code(PyObject *selfPtr, void *closure) {
+  try {
+    py::handle self(selfPtr);
+    py::tuple args = self.attr("args");
+    py::object code = args[1];
+    code.inc_ref();
+    return code.ptr();
+  } catch (py::error_already_set &e) {
+    py::none ret;
+    ret.inc_ref();
+    return ret.ptr();
+  }
+}
+
+static PyGetSetDef TuxedoException_getsetters[] = {
+    {const_cast<char *>("code"), TuxedoException_code, nullptr, nullptr,
+     nullptr},
+    {nullptr}};
+
 struct client {
   client() {}
   // TODO: real params
@@ -59,24 +102,51 @@ struct client {
     if (tpgetctxt(&context, 0) != -1 && context >= 0) {
       return;
     }
+    _init(
+        [client](TPINIT *tpinfo) {
+          if (client) {
+            if (tpinit(tpinfo) == -1) {
+              throw xatmi_exception(tperrno);
+            }
+          } else {
+            if (tpappthrinit(tpinfo) == -1) {
+              throw xatmi_exception(tperrno);
+            }
+          }
+        },
+        nullptr, "tpsysadm", nullptr, nullptr, TPMULTICONTEXTS);
+  }
+  client(std::function<void(TPINIT *)> f, const char *usrname,
+         const char *cltname, const char *passwd, const char *grpname,
+         long flags) {
+    _init(f, usrname, cltname, passwd, grpname, flags);
+  }
 
+  void _init(std::function<void(TPINIT *)> f, const char *usrname,
+             const char *cltname, const char *passwd, const char *grpname,
+             long flags) {
     std::unique_ptr<char, decltype(&tpfree)> guard(
         tpalloc(const_cast<char *>("TPINIT"), nullptr, TPINITNEED(16)),
         &tpfree);
-    TPINIT *tpinitbuf = reinterpret_cast<TPINIT *>(guard.get());
-    memset(tpinitbuf, 0, sizeof(*tpinitbuf));
-    strcpy(tpinitbuf->cltname, "tpsysadm");
-    tpinitbuf->flags = TPMULTICONTEXTS;
-    if (client) {
-      if (tpinit(tpinitbuf) == -1) {
-        throw xatmi_exception(tperrno);
-      }
-    } else {
-      if (tpappthrinit(tpinitbuf) == -1) {
-        throw xatmi_exception(tperrno);
-      }
+    TPINIT *tpinfo = reinterpret_cast<TPINIT *>(guard.get());
+    memset(tpinfo, 0, sizeof(*tpinfo));
+
+    if (usrname != nullptr) {
+      strncpy(tpinfo->usrname, usrname, sizeof(tpinfo->usrname));
     }
+    if (cltname != nullptr) {
+      strncpy(tpinfo->cltname, cltname, sizeof(tpinfo->cltname));
+    }
+    if (passwd != nullptr) {
+      strncpy(tpinfo->passwd, passwd, sizeof(tpinfo->passwd));
+    }
+    if (grpname != nullptr) {
+      strncpy(tpinfo->grpname, grpname, sizeof(tpinfo->grpname));
+    }
+    tpinfo->flags = flags;
+    f(tpinfo);
   }
+
   ~client() {}
   client(const client &) = delete;
   client &operator=(const client &) = delete;
@@ -398,6 +468,19 @@ static py::object pytpimport(const std::string istr, long flags) {
   return to_py(std::move(obuf));
 }
 
+static void pytppost(const std::string eventname, py::object data, long flags) {
+  auto in = from_py(data);
+
+  {
+    py::gil_scoped_release release;
+    int rc =
+        tppost(const_cast<char *>(eventname.c_str()), *in.pp, in.len, flags);
+    if (rc == -1) {
+      throw xatmi_exception(tperrno);
+    }
+  }
+}
+
 static pytpreply pytpcall(const char *svc, py::object idata, long flags) {
   default_client();
   auto in = from_py(idata);
@@ -564,21 +647,18 @@ extern int _tmbuilt_with_thread_option;
 }
 
 static struct tmdsptchtbl_t _tmdsptchtbl[] = {
-    {(char *)"", (char *)"PY", PY, 0, 0}, {NULL, NULL, NULL, 0, 0}};
+    {(char *)"", (char *)"PY", PY, 0, 0}, {nullptr, nullptr, nullptr, 0, 0}};
 
-static struct tmsvrargs_t tmsvrargs = {NULL,        &_tmdsptchtbl[0],
-                                       0,           tpsvrinit,
-                                       tpsvrdone,   _tmrunserver,
-                                       NULL,        NULL,
-                                       NULL,        NULL,
-                                       tprminit,    tpsvrthrinit,
-                                       tpsvrthrdone};
+static struct tmsvrargs_t tmsvrargs = {
+    nullptr,      &_tmdsptchtbl[0], 0,           tpsvrinit, tpsvrdone,
+    _tmrunserver, nullptr,          nullptr,     nullptr,   nullptr,
+    tprminit,     tpsvrthrinit,     tpsvrthrdone};
 
 typedef void *(xao_svc_ctx)(void *);
 static xao_svc_ctx *xao_svc_ctx_ptr;
 struct tmsvrargs_t *_tmgetsvrargs(const char *rmname) {
-  tmsvrargs.reserved1 = NULL;
-  tmsvrargs.reserved2 = NULL;
+  tmsvrargs.reserved1 = nullptr;
+  tmsvrargs.reserved2 = nullptr;
   // tmsvrargs.xa_switch = &xaosw;
   if (strcasecmp(rmname, "NONE") == 0) {
     tmsvrargs.xa_switch = &tmnull_switch;
@@ -630,8 +710,8 @@ PYBIND11_MODULE(tuxedo, m) {
       .def_readonly("rval", &pytpreply::rval)
       .def_readonly("rcode", &pytpreply::rcode)
       .def_readonly("data", &pytpreply::data)
-      .def_readonly("cd", &pytpreply::cd)  // Does not unpack as the use is rare
-                                           // case of tpgetrply(TPGETANY)
+      .def_readonly("cd", &pytpreply::cd)  // Does not unpack as the use is
+                                           // rare case of tpgetrply(TPGETANY)
       .def("__getitem__", [](const pytpreply &s, size_t i) -> py::object {
         if (i == 0) {
           return py::int_(s.rval);
@@ -644,22 +724,47 @@ PYBIND11_MODULE(tuxedo, m) {
         }
       });
 
-  //  py::class_<xatmi_exception>(m, "XatmiException")
-  //      .def_readonly("code", &xatmi_exception::code)
-  //      .def_readonly("message", &xatmi_exception::message);
-  //  py::class_<fml32_exception>(m, "Fml32Exception")
-  //      .def_readonly("code", &fml32_exception::code)
-  //      .def_readonly("message", &fml32_exception::message);
+  static PyObject *XatmiException =
+      PyErr_NewException("tuxedo.XatmiException", nullptr, nullptr);
+
+  if (XatmiException) {
+    PyTypeObject *as_type = reinterpret_cast<PyTypeObject *>(XatmiException);
+    as_type->tp_str = TuxedoException_tp_str;
+    PyObject *descr = PyDescr_NewGetSet(as_type, TuxedoException_getsetters);
+    auto dict = py::reinterpret_borrow<py::dict>(as_type->tp_dict);
+    dict[py::handle(((PyDescrObject *)(descr))->d_name)] = py::handle(descr);
+
+    Py_XINCREF(XatmiException);
+    m.add_object("XatmiException", py::handle(XatmiException));
+  }
+  static PyObject *Fml32Exception =
+      PyErr_NewException("tuxedo.Fml32Exception", nullptr, nullptr);
+  if (Fml32Exception) {
+    PyTypeObject *as_type = reinterpret_cast<PyTypeObject *>(Fml32Exception);
+    as_type->tp_str = TuxedoException_tp_str;
+    PyObject *descr = PyDescr_NewGetSet(as_type, TuxedoException_getsetters);
+    auto dict = py::reinterpret_borrow<py::dict>(as_type->tp_dict);
+    dict[py::handle(((PyDescrObject *)(descr))->d_name)] = py::handle(descr);
+
+    Py_XINCREF(Fml32Exception);
+    m.add_object("Fml32Exception", py::handle(Fml32Exception));
+  }
 
-  static py::exception<xatmi_exception> exc1(m, "XatmiException");
-  static py::exception<fml32_exception> exc2(m, "Fml32Exception");
   py::register_exception_translator([](std::exception_ptr p) {
     try {
-      if (p) std::rethrow_exception(p);
+      if (p) {
+        std::rethrow_exception(p);
+      }
     } catch (const xatmi_exception &e) {
-      exc1(e.what());
+      py::tuple args(2);
+      args[0] = e.what();
+      args[1] = e.code();
+      PyErr_SetObject(XatmiException, args.ptr());
     } catch (const fml32_exception &e) {
-      exc2(e.what());
+      py::tuple args(2);
+      args[0] = e.what();
+      args[1] = e.code();
+      PyErr_SetObject(Fml32Exception, args.ptr());
     }
   });
 
@@ -674,6 +779,63 @@ PYBIND11_MODULE(tuxedo, m) {
       },
       "Returns the OCI service handle for a given XA connection");
 
+  m.def(
+      "tpinit",
+      [](const char *usrname, const char *cltname, const char *passwd,
+         const char *grpname, long flags) {
+        py::gil_scoped_release release;
+        tclient.reset(new client(
+            [](TPINIT *tpinfo) {
+              if (tpinit(tpinfo) == -1) {
+                throw xatmi_exception(tperrno);
+              }
+            },
+            usrname, cltname, passwd, grpname, flags));
+      },
+      "Joins an application", py::arg("usrname") = nullptr,
+      py::arg("cltname") = nullptr, py::arg("passwd") = nullptr,
+      py::arg("grpname") = nullptr, py::arg("flags") = 0);
+
+  m.def(
+      "tpterm",
+      []() {
+        py::gil_scoped_release release;
+        tclient.reset();
+        if (tpterm() == -1) {
+          throw xatmi_exception(tperrno);
+        }
+      },
+      "Leaves an application");
+
+  m.def(
+      "tpappthrinit",
+      [](const char *usrname, const char *cltname, const char *passwd,
+         const char *grpname, long flags) {
+        py::gil_scoped_release release;
+        tclient.reset(new client(
+            [](TPINIT *tpinfo) {
+              if (tpappthrinit(tpinfo) == -1) {
+                throw xatmi_exception(tperrno);
+              }
+            },
+            usrname, cltname, passwd, grpname, flags));
+      },
+      "Joins an application", py::arg("usrname") = nullptr,
+      py::arg("cltname") = nullptr, py::arg("passwd") = nullptr,
+      py::arg("grpname") = nullptr, py::arg("flags") = 0);
+
+  m.def(
+      "tpappthrterm",
+      []() {
+        py::gil_scoped_release release;
+        tclient.reset();
+        if (tpappthrterm() == -1) {
+          throw xatmi_exception(tperrno);
+        }
+      },
+      "Routine for creating and initializing a new Tuxedo context in an "
+      "application-created server thread.");
+
   m.def(
       "tpbegin",
       [](unsigned long timeout, long flags) {
@@ -769,6 +931,9 @@ PYBIND11_MODULE(tuxedo, m) {
         "Converts an exported representation back into a typed message buffer",
         py::arg("istr"), py::arg("flags") = 0);
 
+  m.def("tppost", &pytppost, "Posts an event", py::arg("eventname"),
+        py::arg("data"), py::arg("flags") = 0);
+
   m.def(
       "Fldtype32", [](FLDID32 fieldid) { return Fldtype32(fieldid); },
       "Maps field identifier to field type", py::arg("fieldid"));
@@ -936,6 +1101,8 @@ PYBIND11_MODULE(tuxedo, m) {
 
   m.attr("TPEX_STRING") = py::int_(TPEX_STRING);
 
+  m.attr("TPMULTICONTEXTS") = py::int_(TPMULTICONTEXTS);
+
   m.doc() =
       R"(Python3 bindings for writing Oracle Tuxedo clients and servers
 
@@ -947,7 +1114,7 @@ Flags to service routines:
 - TPNOTRAN - not sent in transaction mode
 - TPTRAN - sent in transaction mode
 - TPNOTIME - no timeout
-- TPABSOLUTE - absolute value on tmsetprio 
+- TPABSOLUTE - absolute value on tmsetprio
 - TPGETANY - get any valid reply
 - TPNOCHANGE - force incoming buffer to match
 - RESERVED_BIT1 - reserved for future use
